"use strict";

var $ = require("../../core/renderer"),
    messageLocalization = require("../../localization/message"),
    extend = require("../../core/utils/extend").extend,
    DataSourceModule = require("../../data/data_source/data_source"),
    deferredUtils = require("../../core/utils/deferred"),
    utils = require("../filter_builder/utils");

function baseOperation(grid) {
    var calculateFilterExpression = function calculateFilterExpression(filterValue, field) {
        var result = [],
            lastIndex = filterValue.length - 1;
        filterValue && filterValue.forEach(function (value, index) {
            if (utils.isCondition(value) || utils.isGroup(value)) {
                var filterExpression = utils.getFilterExpression(value, [field], [], "headerFilter");
                result.push(filterExpression);
            } else {
                result.push(utils.getFilterExpression([field.dataField, "=", value], [field], [], "headerFilter"));
            }
            index !== lastIndex && result.push("or");
        });
        if (result.length === 1) {
            result = result[0];
        }
        return result;
    };

    var getFullText = function getFullText(itemText, parentText) {
        return parentText ? parentText + "/" + itemText : itemText;
    };

    var getSelectedItemsTexts = function getSelectedItemsTexts(items, parentText) {
        var result = [];
        items.forEach(function (item) {
            if (item.items) {
                var selectedItemsTexts = getSelectedItemsTexts(item.items, getFullText(item.text, parentText));
                result = result.concat(selectedItemsTexts);
            }
            item.selected && result.push(getFullText(item.text, parentText));
        });
        return result;
    };

    var headerFilterController = grid && grid.getController("headerFilter"),
        customizeText = function customizeText(fieldInfo) {
        var values = fieldInfo.value || [],
            column = grid.columnOption(fieldInfo.field.dataField),
            headerFilter = column && column.headerFilter,
            lookup = column && column.lookup;

        if (headerFilter && headerFilter.dataSource || lookup && lookup.dataSource) {
            column = extend({}, column, { filterType: "include", filterValues: values });
            var dataSourceOptions = headerFilterController.getDataSource(column);
            dataSourceOptions.paginate = false;
            var dataSource = new DataSourceModule.DataSource(dataSourceOptions),
                result = new deferredUtils.Deferred();

            dataSource.load().done(function (items) {
                texts = getSelectedItemsTexts(items);
                result.resolve(texts.join(", "));
            });
            return result;
        } else {
            var texts = values.map(function (value) {
                var text = headerFilterController.getHeaderItemText(value, column, 0, grid.option("headerFilter"));
                return text;
            });

            return texts.join(", ");
        }
    };
    return {
        dataTypes: ["string", "date", "datetime", "number", "boolean", "object"],
        calculateFilterExpression: calculateFilterExpression,
        editorTemplate: function editorTemplate(conditionInfo, container) {
            var div = $("<div>").addClass("dx-filterbuilder-item-value-text").text(conditionInfo.text).appendTo(container),
                column = extend(true, {}, grid.columnOption(conditionInfo.field.dataField));

            var setValue = function setValue(value) {
                conditionInfo.setValue(value);
            };

            column.filterType = "include";
            column.filterValues = conditionInfo.value ? conditionInfo.value.slice() : [];

            headerFilterController.showHeaderFilterMenuBase({
                columnElement: div,
                column: column,
                apply: function apply() {
                    setValue(this.filterValues);
                    headerFilterController.hideHeaderFilterMenu();
                },
                onHidden: function onHidden() {
                    $("body").trigger("dxpointerdown");
                },
                isFilterBuilder: true
            });
            return container;
        },
        customizeText: customizeText
    };
}

function anyOf(grid) {
    return extend(baseOperation(grid), {
        name: "anyof",
        caption: messageLocalization.format("dxFilterBuilder-filterOperationAnyOf")
    });
}

function noneOf(grid) {
    var baseOp = baseOperation(grid);
    return extend({}, baseOp, {
        calculateFilterExpression: function calculateFilterExpression(filterValue, field) {
            var baseFilter = baseOp.calculateFilterExpression(filterValue, field);
            if (!baseFilter || baseFilter.length === 0) return null;

            return baseFilter[0] === "!" ? baseFilter : ["!", baseFilter];
        },
        name: "noneof",
        caption: messageLocalization.format("dxFilterBuilder-filterOperationNoneOf")
    });
}

exports.anyOf = anyOf;
exports.noneOf = noneOf;